from pqrcore.experiment.signal import Signal, LoopedSignal

import re

from math import inf

import numpy as np

ns = 1.0
us = 1000.0
ms = 1000000.0

MHz = 1.0
kHz = 0.001
Hz = 0.000001


def encode_signal(sig):
    """
        sig - 信号 (Signal类)
    """
    if isinstance(sig, Signal):
        while isinstance(sig, LoopedSignal):
            sig = sig.signal[-1][0]
    else:
        sig = Signal(sig)
    expr = sig.signal.expression
    print(expr)




def is_single_loop(ALoopedSignal):
    simple_bool = True
    if type(ALoopedSignal) == LoopedSignal:
        for i in range(len(ALoopedSignal.signal)):
            k = ALoopedSignal.signal[i]
            if type(k[0]) != Signal:  # 判断是否为简单循环
                simple_bool = False
                break
    elif type(ALoopedSignal) == np.ndarray:
        for i in range(ALoopedSignal.shape[0]):
            if ALoopedSignal[i, 2] != 0:  # 判断是否为简单循环
                simple_bool = False
                break
    else:
        raise ValueError("传参类型不对！")
        return -1
    return simple_bool


def decode_simple_loop(ALoopedSignal):
    """
    处理简单循环
        sequence1[on/off , duration, inst, inst_data] - 单通道序列
        lasting_time - 循环一次持续时间
        loop_times - 循环次数
    :param ALoopedSignal:
    :return:
    """
    if type(ALoopedSignal) != LoopedSignal:
        raise ValueError("传参类型不对！")

    if len(ALoopedSignal.signal) == 1:
        sequence = np.array([ALoopedSignal.signal[0][0].signal.expression,  # 将第一行写入序列
                              ALoopedSignal.signal[0][1], -1, 0])
        lasting_time = ALoopedSignal.signal[0][1]
        loop_times = 0
        return sequence, lasting_time, loop_times

    sequence1 = np.array([ALoopedSignal.signal[0][0].signal.expression,  # 将第一行写入序列
                          ALoopedSignal.signal[0][1]])
    if len(ALoopedSignal.signal) == 2:
        temp = np.array([ALoopedSignal.signal[1][0].signal.expression,  # 将第二行写入序列
                         ALoopedSignal.signal[1][1] - ALoopedSignal.signal[0][1]])
        sequence1 = np.vstack((sequence1, temp))
    else:
        for i in range(len(ALoopedSignal.signal) - 1):
            temp = np.array([ALoopedSignal.signal[i + 1][0].signal.expression,  # 将第一行写入序列
                             ALoopedSignal.signal[i + 1][1] - ALoopedSignal.signal[i][1]])
            sequence1 = np.vstack((sequence1, temp))
    status = np.zeros((len(ALoopedSignal.signal), 2))
    status[0][0] = 2  # 声明循环开始
    status[0][1] = ALoopedSignal.loops  # 声明循环次数
    status[-1][0] = 3  # 声明循环结束
    sequence1 = np.hstack((sequence1, status))  # 合并sequence和status
    lasting_time = ALoopedSignal.signal[-1][1]
    loop_times = ALoopedSignal.loops
    return sequence1, lasting_time, loop_times


def adjust_limit_loop(ALoopedSignal, limited_time):
    """
    处理有时长限制的简单循环
    :param ALoopedSignal:可以是signal类,也可以是np.array类
    :param limited_time: 循环的限制时间
    :return:
    """
    if type(ALoopedSignal) == LoopedSignal:
        simple_sequence, lasting_time, loop_times = decode_simple_loop(ALoopedSignal)
    elif type(ALoopedSignal) == np.ndarray:
        simple_sequence = ALoopedSignal
        lasting_time = sum(ALoopedSignal[:, 1])
        loop_times = ALoopedSignal[0, 3]
    else:
        raise ValueError("传参类型不对！")
        return -1

    if lasting_time*loop_times == limited_time:
        return simple_sequence

    elif lasting_time*loop_times > limited_time:  # 循环时长超过限制时长，需要缩短序列
        mismatch_time = lasting_time*loop_times - limited_time
        reduced_times = 1 + mismatch_time // lasting_time
        if loop_times <= reduced_times:
            raise ValueError("时序长度不对!")
        else:
            simple_sequence[0, 3] -= reduced_times
            mismatch_time = limited_time - limited_time // lasting_time * lasting_time
            temp = 0
            add_sequence = np.array([0, 0, 0, 0])
            for i in range(simple_sequence.shape[0]):
                temp += simple_sequence[i][1]
                if temp <= mismatch_time:
                    inst = np.array([simple_sequence[i][0], simple_sequence[i][1], 0, 0])
                    add_sequence = np.vstack((add_sequence, inst))
                else:
                    inst = np.array([simple_sequence[i][0], mismatch_time - (temp - simple_sequence[i][1]), 0, 0])
                    add_sequence = np.vstack((add_sequence, inst))
                    break
            simple_sequence = np.vstack((simple_sequence, add_sequence[1:, :]))

            lens = len(simple_sequence)
            for j in range(lens):
                t = lens - j - 1
                if simple_sequence[t, 1] == 0:
                    simple_sequence = np.delete(simple_sequence, t, axis=0)  # 去除序列中时间为0的部分
            return simple_sequence

    else:  # 循环时长小于限制时长，需要扩增序列
        mismatch_time = abs(lasting_time*loop_times - limited_time)
        zero_sequence = np.array([0, mismatch_time, 0, 0])
        simple_sequence = np.vstack((simple_sequence, zero_sequence))
        return simple_sequence


def combine_similar_items(sequence):
    '''
    合并冗余项
    :param sequence: 未处理序列
    :return:
    '''
    if len(sequence.shape) ==1:
        return sequence

    while True:  # 合并冗余语句
        have_similar_items = False
        for i in range(sequence.shape[0] - 1):  # 全部遍历，检测是否能合并语句
            if sequence[i, 0] == sequence[i + 1, 0] and sequence[i, 2] == sequence[i + 1, 2] == 0:
                have_similar_items = True

        if not have_similar_items:
            break

        for i in range(sequence.shape[0] - 1):
            if sequence[i, 0] == sequence[i + 1, 0] and sequence[i, 2] == sequence[i + 1, 2] == 0:
                sequence[i, 1] += sequence[i + 1, 1]
                sequence = np.delete(sequence, i + 1, axis=0)
                break
    lens = sequence.shape[0]
    for j in range(lens):
        i = lens - j - 1
        if sequence[i, 1] == 0:
            sequence = np.delete(sequence, i, axis=0)  # 删除无效声明
    if sequence[0, 1] == 0:
        sequence = np.delete(sequence, 0, axis=0)
    return sequence


def edge_detect(f):
    '''
    :param f:
    :return:
    '''
    N = len(f)
    edge = np.zeros((N - 1))
    for i in range(N - 1):
        edge[i] = f[i] - f[i + 1]
    return edge


def adjust_loop_delay(loop_sequence, delay):
    """
    考虑延迟后将简单循环进行错位
    默认响应延迟时间要小于单次循环时间！
    :param loop_sequence: 需要错位的简单循环
    :param delay: 本通道的响应延迟
    :return: adjusted_sequence: 调整后的时间序列
    """
    if delay == 0:
        return loop_sequence
    loop_times = loop_sequence[0, 3]
    front_sequence = np.array([0, 0])
    temp = 0
    middle_sequence = loop_sequence[:, 0:2]
    end_sequence = loop_sequence[:, 0:2]
    for i in range(loop_sequence.shape[0]):
        temp += loop_sequence[i, 1]
        if temp <= delay:
            insert_sequence = loop_sequence[i, 0:2]
            # print(front_sequence, insert_sequence)
            front_sequence = np.vstack((front_sequence, insert_sequence))
            middle_sequence = np.delete(middle_sequence, 0, axis=0)
            middle_sequence = np.vstack((middle_sequence, insert_sequence))
            end_sequence = np.delete(end_sequence, 0, axis=0)
        else:
            if temp == loop_sequence[i-1, 1]+delay:
                break
            mismatch_time = temp - delay
            insert_sequence = np.array([loop_sequence[i, 0], delay])
            complementary_sequence = np.array([loop_sequence[i, 0], mismatch_time])
            front_sequence = np.vstack((front_sequence, insert_sequence))
            middle_sequence = np.delete(middle_sequence, 0, axis=0)
            middle_sequence = np.vstack((complementary_sequence, middle_sequence))
            middle_sequence = np.vstack((middle_sequence, insert_sequence))
            end_sequence = np.delete(end_sequence, 0, axis=0)
            end_sequence = np.vstack((complementary_sequence, end_sequence))
            break

    front_status = np.zeros((front_sequence.shape[0], 2))
    middle_status = np.zeros((middle_sequence.shape[0], 2))
    end_status = np.zeros((end_sequence.shape[0], 2))

    front_sequence = np.hstack((front_sequence, front_status))
    middle_sequence = np.hstack((middle_sequence, middle_status))
    end_sequence = np.hstack((end_sequence, end_status))

    middle_sequence[0, 2] = 2
    middle_sequence[0, 3] = loop_times - 1
    middle_sequence[-1, 2] = 3

    adjusted_sequence = np.vstack((front_sequence, middle_sequence, end_sequence))
    adjusted_sequence = combine_similar_items(adjusted_sequence)
    # print('front_sequence', front_sequence)
    # print("middle_sequence", middle_sequence)
    # print('end_sequence', end_sequence)
    return adjusted_sequence


def adjust_simple_delay(simple_sequence, delay):
    '''
    处理简单序列的延迟问题
    :param simple_sequence: 简单序列
    :param delay: 延迟
    :return:
    '''
    if delay == 0:
        return simple_sequence
    lens = simple_sequence.shape[0]
    edge = edge_detect(simple_sequence[:, 0])
    insert_number = 0
    for i in range(lens - 1):
        if edge[i] == -1:
            temp = np.array([0, edge[i] * delay, 0, 0])
            simple_sequence = np.insert(simple_sequence, i + insert_number + 1, temp, axis=0)
            insert_number += 1
        elif edge[i] == 1:
            temp = np.array([0, edge[i] * delay, 0, 0])
            simple_sequence = np.insert(simple_sequence, i + insert_number + 1, temp, axis=0)
            insert_number += 1
    simple_sequence = combine_similar_items(simple_sequence)
    return simple_sequence


def adjust_general_delay(sequence, delay, max_delay):
    if len(sequence.shape) == 1:
        sequence[1] += max_delay
        sequence = combine_similar_items(sequence)
        return sequence
    sequence = np.vstack((np.array([0, max_delay, 0, 0]), sequence))
    if delay == 0:
        sequence = combine_similar_items(sequence)
        return sequence

    index_split = []  # 切割序列

    sequence_start = 0
    sequence_end = 0
    loop_start = 0
    loop_end = 0
    state = sequence[:, 2]


    if is_single_loop(sequence):
        index_split.append([0, sequence.shape[0], "simple_sequence"])
    else:
        for i in range(sequence.shape[0]):
            if state[i] == 2:
                loop_start = i
                sequence_end = i - 1
                index_split.append([sequence_start, sequence_end, "simple_sequence"])
            if state[i] == 3:
                loop_end = i
                sequence_start = i + 1
                index_split.append([loop_start, loop_end, "loop"])
        if loop_end < len(state) - 1:
            index_split.append([loop_end + 1, len(state) - 1, "simple_sequence"])
    new_sequence = np.array([0, 0, 0, 0])

    for x in index_split:
        temp = sequence[x[0]:x[1]+1, :]
        if x[2] == "simple_sequence":
            if x[0] != 0:
                temp = np.vstack((np.array([1, 0, 0, 0]), temp))
            if x[1] != len(state):
                temp = np.vstack((temp, np.array([1, 0, 0, 0])))
            # print(adjust_simple_delay(temp, delay))
            new_sequence = np.vstack((new_sequence, adjust_simple_delay(temp, delay)))
        if x[2] == "loop":
            # print(adjust_loop_delay(temp, delay))
            new_sequence = np.vstack((new_sequence, adjust_loop_delay(temp, delay)))
    new_sequence = np.vstack((new_sequence, np.array([0, delay, 0, 0])))  # 补足时序长度
    sequence = combine_similar_items(new_sequence)

    return sequence


def decode_multi_loop(ALoopedSignal):
    """
    大循环中有多个一级小循环
    :param ALoopedSignal: 两层嵌套循环
    :return: sequence: 两层单通道序列(不考虑延迟)
    """
    if len(ALoopedSignal.signal) == 1:
        sequence = decode_simple_loop(ALoopedSignal)[0]
        return sequence
    sequence = np.array([0, 0, 0, 0])  # 空白序列
    for i in range(len(ALoopedSignal.signal)):  # 解包双层循环
        k = ALoopedSignal.signal[i]
        if i == 0:
            begin_time = 0
        else:
            p = ALoopedSignal.signal[i - 1]
            begin_time = p[1]  # 上次状态的结束时间
        if type(k[0]) == Signal:
            temp = np.array([k[0].signal.expression, k[1] - begin_time, 0, 0])
            sequence = np.vstack((sequence, temp))
            # print('Signal', temp)
        else:
            temp = adjust_limit_loop(k[0], k[1] - begin_time)
            sequence = np.vstack((sequence, temp))
            # print('Loop', k[0], k[1] - begin_time)
    sequence = np.delete(sequence, 0, axis=0)  # 删除补位空序列

    while True:  # 展开小于等于两次的循环
        have_useless_loop = False
        for i in range(sequence.shape[0]):  # 全部遍历，检测是否有小于两次的循环
            if 0 < sequence[i, 3] < 3:
                have_useless_loop = True

        if not have_useless_loop:
            break

        for i in range(sequence.shape[0]):
            if 0 < sequence[i, 3] < 3:
                loop_times = sequence[i, 3]
                find_min = True
                nearest_index = 0
                while find_min:
                    if sequence[i + nearest_index, 2] == 3:
                        find_min = False
                    else:
                        nearest_index += 1
                if loop_times == 1:
                    sequence[i, 3] = 0
                    sequence[i, 2] = 0
                    sequence[i + nearest_index, 2] = 0
                else:
                    sequence[i, 3] = 0
                    sequence[i, 2] = 0
                    sequence[i + nearest_index, 2] = 0
                    for k in range(nearest_index + 1):
                        sequence = np.insert(sequence, i + nearest_index + k + 1, sequence[i + k, :], axis=0)
                break
    sequence = combine_similar_items(sequence)  # 合并同类项
    sequence[0, 2] = 2
    sequence[-1, 2] = 3
    sequence[0, 3] = ALoopedSignal.loops

    return sequence


def decode_recursion_loop(ALoopedSignal):
    """
    未开发成功
    :param ALoopedSignal:
    :return:
    """
    # for k in ALoopedSignal._sig:
    #     if type(k[0]) == Signal:
    #         print(str(type(k[0]))+" depth = " + str(depth))
    #         end_time = k[1]
    #         previous_index = ALoopedSignal._sig.index(k)-1
    #         if previous_index < 0:
    #             start_time = 0
    #         else:
    #             start_time = ALoopedSignal._sig[previous_index][1]
    #         print(end_time-start_time)
    #     else:
    #         decode_recursion_loop(k[0], depth+1)  # 递归,深度加一
    #     if ALoopedSignal._sig.index(k) == len(ALoopedSignal._sig)-1:  # 声明Loop结束
    #         print("End Loop")
    if is_single_loop(ALoopedSignal):
        return decode_simple_loop(ALoopedSignal)[0]
    else:
        empty_sequence = np.array([0, 0, 0, 0])
        for i in range(len(ALoopedSignal.signal)):
            k = ALoopedSignal.signal[i]
            if i == 0:
                begin_time = 0
            else:
                p = ALoopedSignal.signal[i - 1]
                begin_time = p[1]  # 上次状态的结束时间
            limit_time = k[1] - begin_time
            if type(k[0]) == Signal:
                temp = np.array([k[0].signal.expression, limit_time, 0, 0])
                empty_sequence = np.vstack((empty_sequence, temp))
            else:
                return empty_sequence
        return empty_sequence


def synchronize_simple_sqeuence(sequence_list):
    """

    :param sequence_list: [[port_1,sequence_1]...[port_n,sequence_n]]
    :return:
    """
    timeline = [0]
    for channel_index, channel in enumerate(sequence_list):
        if len(channel[1].shape) == 1:
            timeline.append(channel[1][1])
            continue
        else:
            for index, instruction in enumerate(channel[1]):
                timeline.append(sum(channel[1][0:index, 1]))
        timeline.append(sum(channel[1][:, 1]))
    timeline = list(sorted(set(timeline)))
    cell_line = [timeline[i+1] - timeline[i] for i in range(len(timeline)-1)]
    all_channels = np.zeros((len(cell_line), 4))
    all_channels[:, 1] = cell_line

    for channel_index, channel in enumerate(sequence_list):
        port = channel[0]
        sequence = channel[1]
        split_sequence = []

        index = 0
        stat_index = 0
        sub_index = 0
        # print(sequence[:, 1])
        if len(sequence.shape) == 1:
            bin_list = ["0" for i in range(port-1)]
            bin_list.insert(0, "0b1")
            bin_str = "".join(bin_list)
            all_channels[:, 0] += int(bin_str, 2)
            continue
        while stat_index < sequence.shape[0]:
            part = []
            temp = 0
            while True:
                if sub_index + index == len(cell_line):
                    break
                temp += cell_line[index+sub_index]
                part.append(cell_line[index+sub_index])
                if temp < sequence[stat_index][1]:
                    sub_index += 1
                else:
                    split_sequence.append([sequence[stat_index][0], part])
                    break
            index += sub_index+1
            stat_index += 1
            sub_index = 0

        index = 0
        cell_index = 0
        while index < len(split_sequence):
            for i in range(len(split_sequence[index][1])):
                bin_list = ["0" for j in range(port - 1)]
                bin_list.insert(0, "0b"+str(int(split_sequence[index][0])))
                bin_str = "".join(bin_list)
                all_channels[cell_index, 0] += int(bin_str, 2)
                cell_index += 1
            index += 1

        # print("split_sequence", split_sequence)

    return all_channels


def synchronize_complex_sqeuence(sequence_list):
    """

    :param sequence_list: [[port_1,sequence_1]...[port_n,sequence_n]]
    :return:
    """
    part = [[] for i in sequence_list]
    all_channels = np.zeros((1, 4))
    special_channel = []
    normal_channel = []
    for index, ch in enumerate(sequence_list):
        index_split = part[index]  # 切割序列
        sequence = ch[1]
        if len(sequence.shape) == 1:
            special_channel.append(index)
            continue
        else:
            normal_channel.append(index)
        sequence_start = 0
        sequence_end = 0
        loop_start = 0
        loop_end = 0
        state = sequence[:, 2]
        if is_single_loop(sequence):
            index_split.append([0, sequence.shape[0], "simple_sequence"])
        else:
            for i in range(sequence.shape[0]):
                if state[i] == 2:
                    loop_start = i
                    sequence_end = i - 1
                    index_split.append([sequence_start, sequence_end, "simple_sequence"])
                if state[i] == 3:
                    loop_end = i
                    sequence_start = i + 1
                    index_split.append([loop_start, loop_end, "loop"])
            if loop_end < len(state) - 1:
                index_split.append([loop_end + 1, len(state) - 1, "simple_sequence"])
        # print(index_split)

    for i in range(len(part[0])):
        temp = []
        for k in range(len(part)):
            if k in special_channel:
                continue
            if len(sequence_list[k][1]) == 1:
                temp.append([sequence_list[k][0], sequence_list[k][1][0, :]])
            else:
                temp.append([sequence_list[k][0], sequence_list[k][1][part[k][i][0]:part[k][i][1]+1, :]])
        synchronized_sequence = synchronize_simple_sqeuence(temp)
        # print("synchronized_sequence", synchronized_sequence)
        if part[0][i][2] == 'loop':
            normal_index = normal_channel[0]
            synchronized_sequence[0, 2] = 2
            synchronized_sequence[0, 3] = sequence_list[normal_index][1][part[normal_index][i][0], 3]
            synchronized_sequence[-1, 2] = 3
        all_channels = np.vstack((all_channels, synchronized_sequence))
        # print("finish one part")

    for i in special_channel:
        port = sequence_list[i][0]
        bin_list = ["0" for i in range(port - 1)]
        bin_list.insert(0, "0b"+str(int(sequence_list[i][1][0])))
        bin_str = "".join(bin_list)
        all_channels[:, 0] += int(bin_str, 2)
    all_channels = combine_similar_items(all_channels)
    # print("all_channels", all_channels)
    return all_channels




import os

from chardet import detect

# fileSuffix = input("请输入文件后缀")


def convert_pb(fileSuffix):

    fns = []
    filedir = os.path.join(os.path.abspath('.'), "")
    file_name = os.listdir(os.path.join(os.path.abspath('.'), ""))
    for fn in file_name:
        if fn.endswith(fileSuffix):  # 这里填文件后缀
            fns.append(os.path.join(filedir, fn))

    for fn in fns:
        with open(fn, 'rb+') as fp:
            content = fp.read()
            codeType = detect(content)['encoding']
            content = content.decode(codeType, "ignore").encode("utf8")
            fp.seek(0)
            fp.write(content)
            print(fn, "：已修改为utf8编码")


# -*- coding: utf-8 -*-
"""
Created on 2020-04-21

@author: 张宽

信号发生器
"""

from devices.device import Channel, Device
from experiment.signal import Signal


class GeneratorChannel(Channel):
    """
    发生器通道基础类
    self._sig: 已加载的信号
    """

    @property
    def signal(self):
        try:
            sig = self._sig
        except AttributeError:
            sig = Signal(0)
        return sig

    @signal.setter
    def signal(self, sig):
        if not isinstance(sig, Signal):
            sig = Signal(sig)
        self._sig = sig


class Generator(Device):
    """发生器基础类"""
    _ch_type = GeneratorChannel

    @property
    def signals(self):
        sigs = {}
        for ch in self:
            sigs[ch] = ch.signal
        return sigs

    @signals.setter
    def signals(self, sigs):
        sig = Signal(0)
        for ch in self:
            ch.signal = sigs.get(ch, sig)



# -*- coding: utf-8 -*-
"""
Created on 2020-04-21

@author: 张宽

设备包括发生器、探测器、监视器等
"""

import weakref


class Channel():
    """
    通道基础类
    self.__dev - 从属设备的弱引用 (weakref类)
    self.__ch - 通道信息，如编号 (int类)、描述 (str类)
    """

    def __new__(cls, dev, ch):
        """通道只能由其从属设备创建"""
        if dev in dev.gets():
            return dev[ch]
        self = super().__new__(cls)
        self.__dev = weakref.ref(dev)
        self.__ch = ch
        return self

    def __str__(self):
        return '%s[%s]' % (self.device, self.__ch)

    @property
    def device(self):
        return self.__dev()

    @property
    def channel(self):
        return self.__ch


class Device():
    """
    设备基础类
    cls._devs = {(brand, model, sn): dev, ...} - 设备注册表
         brand - 厂商 (str类)
        model - 型号 (str类)
        sn - 序列号 (str类)
        dev - 设备 (Device类)
    cls._ch_type - 设备通道类型 (type类 = Channel或其子类)
    cls._chs = [ch, ...]
        ch - 通道信息，如编号 (int类)、描述 (str类)
    self.__brand - 厂商 (str类)
    self.__model - 型号 (str类)
    self.__sn - 序列号 (str类)
    self.__chs = {ch: chan, ...} - 全部通道
        ch - 通道信息，如编号 (int类)、描述 (str类)
        chan - 通道 (Channel类)
    self.__name - 设备名 (str类)
    self._handle (int类)
    """
    _devs = {}
    _ch_type = Channel
    _chs = []

    def __new__(cls, brand, model, sn, name='', handle=-1):
        devs = Device._devs
        brand = str(brand)
        model = str(model)
        sn = str(sn)
        try:
            self = devs[brand, model, sn]
        except KeyError:
            self = super().__new__(cls)
            self.__brand = brand
            self.__model = model
            self.__sn = sn
            self.__chs = chs = {}
            for ch in cls._chs:
                chs[ch] = self._ch_type(self, ch)
            self._handle = int(handle)
            devs[brand, model, sn] = self
        else:
            if type(self) != cls:
                raise TypeError('%s不是%s类' % (self, cls.__name__))
            if self._handle != handle:
                raise KeyError('%s的%s序列号重复' % (brand, model))
        self.name = name
        return self

    def __str__(self):
        if self.__name == '':
            return '%s_%s_%s' % (self.__brand, self.__model, self.__sn)
        return self.__name

    def __len__(self):
        return len(self.__chs)

    def __getitem__(self, ch):
        return self.__chs[ch]

    def __iter__(self):
        return iter(self.__chs.values())

    def __contains__(self, chan):
        return chan in self.__chs.values()

    @property
    def brand(self):
        return self.__brand

    @property
    def model(self):
        return self.__model

    @property
    def serial_number(self):
        return self.__sn

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        self.__name = str(name)

    @classmethod
    def get(cls, *args):
        """返回指定设备"""
        return cls._devs[args]

    @classmethod
    def gets(cls):
        """返回全部设备列表"""
        return list(cls._devs.values())

    def delete(self):
        """删除设备"""
        brand = self.__brand
        model = self.__model
        sn = self.__sn
        del Device._devs[brand, model, sn]



import numpy as np


import pandas as pd


import os



def data_trans(filename):
    try:
        origin_data = pd.read_excel(filename, encoding='utf-8')
        del origin_data['Name']
        origin_data.replace(np.nan, 0, inplace=True)
        origin_data = origin_data.values
    except IOError:
        print("Error: 没有找到文件或读取文件失败")
        os._exit(0)

    new_data = origin_data[:, 0:3]
    for k in range(int((origin_data.shape[1] - 3) / 2)):
        new_data = np.row_stack((new_data, np.column_stack((origin_data[:, 0], origin_data[:, 2 * k + 3:2 * k + 5]))))

    start_time = new_data[:, 1].flatten()
    end_time = new_data[:, 2].flatten() + start_time
    port = new_data[:, 0].flatten()
    port_loop = len(port)
    time_dots = np.array(list(sorted(set(np.concatenate((end_time, start_time), axis=0)))))

    Statement_number = len(time_dots) - 1
    Port_number = 24

    time_interval = [time_dots[k + 1] - time_dots[k] for k in range(Statement_number)]
    time_interval = np.array(time_interval)
    time_sequence = np.zeros((Statement_number, Port_number + 2))
    time_sequence[:, -1] = time_interval
    time_sequence[:, 0] = time_dots[:-1] + 1

    # 通道处理
    for k in range(Statement_number):
        for n in range(port_loop):
            if time_sequence[k, 0] > start_time[n] and time_sequence[k, 0] < end_time[n]:
                time_sequence[k, int(port[n])] = 1

    index_of_final = list(np.where(end_time == max(end_time))[0])

    time_sequence = time_sequence[:, 1:]

    time_sequence_new = []
    for n in range(Statement_number):
        temp_list = [int(x) for x in time_sequence[n, :-1].tolist()]
        temp_list = [str(x) for x in temp_list]
        temp = hex(int(''.join(temp_list), 2))
        time_sequence_new.append([temp, time_sequence[n, -1]])

    return time_sequence_new


def data_write(filename, data):
    with open(filename, "w") as f:
        for item in data:
            # f.write(str(bin(int(item[0], 16)))+','+str(item[1])+'ns')
            f.write(item[0] + ',' + str(item[1]) + 'ns')
            f.write('\n')


# origin_time_data = 'loop.xlsx'
#
# time_sequence = data_trans(origin_time_data)
# data_write('loop.pb', time_sequence)
# print(time_sequence)








# -*- coding: utf-8 -*-
"""
Created on 2020-04-02

@author: 张宽

将字符串表达式解析为S-表达式
将S-表达式反解析为字符串表达式
简化、计算S-表达式
"""

from math import *
from operator import *
import ast


def isname(name):
    try:
        name = ast.parse(name)
    except SyntaxError:
        return False
    name = name.body
    if len(name) != 1:
        return False
    name = name[0]
    if not isinstance(name, ast.Expr):
        return False
    name = name.value
    return isinstance(name, ast.Name)


class ParsedExpression():
    """
    S-表达式
    cls.__func_dict = {func: (op, prec), ...}
        func - 操作符对应函数名 (str类)
        op - 操作符 (str类)
        prec - 操作优先级
    self.__expr = (func, *args)
        func - 函数名 (str类)
        arg - 参数 (分形结构，底层为int,float,complex的数字，或str的变量)
    """
    __func_dict = {
        'or_': ('|', 0),
        'xor': ('^', 1),
        'and_': ('&', 2),
        'lshift': ('<<', 3),
        'rshift': ('>>', 3),
        'add': ('+', 4),
        'sub': ('-', 4),
        'mul': ('*', 5),
        'truediv': ('/', 5),
        'floordiv': ('//', 5),
        'mod': ('%', 5),
        'pos': ('+', 6),
        'neg': ('-', 6),
        'invert': ('~', 6),
        'pow': ('**', 7)}

    def __init__(self, expr):
        def parse_ast(expr):
            """将AST表达式解析为S-表达式"""
            # BinOp类型 (e.g. 1 + 2 -> ('+', 1, 2))
            if isinstance(expr, ast.BinOp):
                op_dict = {
                    ast.Add: 'add',
                    ast.Sub: 'sub',
                    ast.Mult: 'mul',
                    ast.Div: 'truediv',
                    ast.Mod: 'mod',
                    ast.Pow: 'pow',
                    ast.LShift: 'lshift',
                    ast.RShift: 'rshift',
                    ast.BitOr: 'or_',
                    ast.BitXor: 'xor',
                    ast.BitAnd: 'and_',
                    ast.FloorDiv: 'floordiv'}
                left = parse_ast(expr.left)
                right = parse_ast(expr.right)
                op = op_dict[type(expr.op)]
                return op, left, right
            # UnaryOp类型 (e.g. -x -> ('-', 'x'))
            if isinstance(expr, ast.UnaryOp):
                op_dict = {
                    ast.Invert: 'invert',
                    ast.UAdd: 'pos',
                    ast.USub: 'neg'}
                operand = parse_ast(expr.operand)
                op = op_dict[type(expr.op)]
                return op, operand
            # Call类型 (e.g. cos(pi) -> ('cos', 'pi'))
            if isinstance(expr, ast.Call):
                args = [parse_ast(arg) for arg in expr.args]
                func = expr.func.id
                return (func, *args)
            # Num类型 (e.g. 0 -> 0)
            if isinstance(expr, ast.Num):
                n = expr.n
                return n
            # Name类型 (e.g. x -> 'x')
            if isinstance(expr, ast.Name):
                id = expr.id
                return id
            # 其它
            raise SyntaxError('AST表达式解析失败')

        if isinstance(expr, str):
            expr = ast.parse(expr).body
            err = SyntaxError('非数学表达式')
            if len(expr) != 1:
                raise err
            expr = expr[0]
            if not isinstance(expr, ast.Expr):
                raise err
            expr = expr.value
            self.__expr = parse_ast(expr)
            return
        self.expression = expr

    def __str__(self):
        def deparse(expr):
            """将S-表达式反解析为(字符串, 函数优先级)"""
            if not isinstance(expr, tuple):
                return str(expr), 8
            func, *args = expr
            # 0个参数
            if args == ():
                return '%s()' % func, 8
            func, prec = self.__func_dict.get(func, (func, -1))
            arg, *args = args
            arg, prec_child = deparse(arg)
            if prec_child < prec:
                arg = '(%s)' % arg
            # 1个参数
            if args == ():
                # 函数不能化为运算符
                if prec < 0:
                    expr = '%s(%s)' % (func, arg)
                    prec = 8
                # 函数能化为运算符
                else:
                    expr = func + arg
                return expr, prec
            expr = arg
            # 函数不能化为运算符
            if prec < 0:
                for arg in args:
                    arg = deparse(arg)[0]
                    expr += ',%s' % arg
                expr = '%s(%s)' % (func, expr)
                prec = 8
            # 函数能化为运算符
            else:
                for arg in args:
                    arg, prec_child = deparse(arg)
                    if prec_child <= prec:
                        arg = '(%s)' % arg
                    expr += func + arg
            return expr, prec

        return deparse(self.__expr)[0]

    def __call__(self, **rules):
        def eval(expr, **rules):
            """计算S-表达式"""
            if not isinstance(expr, tuple):
                if isinstance(expr, str):
                    rules = {**globals(), **rules}
                    expr = rules[expr]
                    if not isinstance(expr, (int, float, complex)):
                        raise TypeError('表达式底层参数应为数字')
                return expr
            func, *args = expr
            args = [eval(arg, **rules) for arg in args]
            rules = {**globals(), **rules}
            func = rules[func]
            return func(*args)

        return eval(self.__expr, **rules)

    def __eq__(self, other):
        if not isinstance(other, ParsedExpression):
            return False
        other = self.__simplify(other.__expr)
        return self.__simplify(self.__expr) == other

    def __pos__(self):
        expr = self.__new__(ParsedExpression)
        expr.__expr = ('pos', self.__expr)
        return expr

    def __neg__(self):
        expr = self.__new__(ParsedExpression)
        expr.__expr = ('neg', self.__expr)
        return expr

    def __invert__(self):
        expr = self.__new__(ParsedExpression)
        expr.__expr = ('invert', self.__expr)
        return expr

    def __add__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('add', self.__expr, other.__expr)
        return expr

    def __radd__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('add', other.__expr, self.__expr)
        return expr

    def __iadd__(self, other):
        if not isinstance(other, ParsedExpression):
            other = ParsedExpression(other)
        self.__expr = ('add', self.__expr, other.__expr)
        return self

    def __sub__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('sub', self.__expr, other.__expr)
        return expr

    def __rsub__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('sub', other.__expr, self.__expr)
        return expr

    def __isub__(self, other):
        if not isinstance(other, ParsedExpression):
            other = ParsedExpression(other)
        self.__expr = ('sub', self.__expr, other.__expr)
        return self

    def __mul__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('mul', self.__expr, other.__expr)
        return expr

    def __rmul__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('mul', other.__expr, self.__expr)
        return expr

    def __imul__(self, other):
        if not isinstance(other, ParsedExpression):
            other = ParsedExpression(other)
        self.__expr = ('mul', self.__expr, other.__expr)
        return self

    def __truediv__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('truediv', self.__expr, other.__expr)
        return expr

    def __rtruediv__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('truediv', other.__expr, self.__expr)
        return expr

    def __itruediv__(self, other):
        if not isinstance(other, ParsedExpression):
            other = ParsedExpression(other)
        self.__expr = ('truediv', self.__expr, other.__expr)
        return self

    def __floordiv__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('floordiv', self.__expr, other.__expr)
        return expr

    def __rfloordiv__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('floordiv', other.__expr, self.__expr)
        return expr

    def __ifloordiv__(self, other):
        if not isinstance(other, ParsedExpression):
            other = ParsedExpression(other)
        self.__expr = ('floordiv', self.__expr, other.__expr)
        return self

    def __mod__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('mod', self.__expr, other.__expr)
        return expr

    def __rmod__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('mod', other.__expr, self.__expr)
        return expr

    def __imod__(self, other):
        if not isinstance(other, ParsedExpression):
            other = ParsedExpression(other)
        self.__expr = ('mod', self.__expr, other.__expr)
        return self

    def __pow__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('pow', self.__expr, other.__expr)
        return expr

    def __rpow__(self, other):
        if isinstance(other, ParsedExpression):
            expr = self.__new__(ParsedExpression)
        else:
            expr = other = ParsedExpression(other)
        expr.__expr = ('pow', other.__expr, self.__expr)
        return expr

    def __ipow__(self, other):
        if not isinstance(other, ParsedExpression):
            other = ParsedExpression(other)
        self.__expr = ('pow', self.__expr, other.__expr)
        return self

    @property
    def expression(self):
        return self.__expr

    @expression.setter
    def expression(self, expr):
        def check(expr):
            """检查S-表达式"""
            if not isinstance(expr, tuple):
                if isinstance(expr, (int, float, complex)):
                    return
                if isinstance(expr, str):
                    if isname(expr):
                        return
                    raise SyntaxError('表达式底层字符串应为变量')
                raise TypeError('表达式底层参数应为数字或字符串')
            func, *args = expr
            if not isinstance(func, str):
                raise TypeError('表达式函数名应为字符串')
            if not isname(func):
                raise SyntaxError('表达式函数名不合法')
            for arg in args:
                check(arg)

        check(expr)
        self.__expr = expr

    @staticmethod
    def __simplify(expr):
        """化简S-表达式"""
        if not isinstance(expr, tuple):
            if isinstance(expr, str):
                try:
                    val = globals()[expr]
                except KeyError:
                    pass
                else:
                    if isinstance(val, (int, float, complex)):
                        expr = val
            return expr
        args = []
        isnum = True
        for arg in expr[1:]:
            arg = ParsedExpression.__simplify(arg)
            args.append(arg)
            if isinstance(arg, (str, tuple)):
                isnum = False
        func = expr[0]
        if isnum:
            try:
                val = globals()[func]
            except KeyError:
                pass
            else:
                val = val(*args)
                if isinstance(val, (int, float, complex)):
                    return val
        return (func, *args)

    def simplify(self):
        """化简self.expression"""
        self.__expr = self.__simplify(self.__expr)



import ctypes

import re

import pqrspincore.DecodeSignal as ds


PULSE_PROGRAM = 0
FREQ_REGS = 1

try:
    spinapi = ctypes.CDLL("spinapi64")
except:
    try:
        spinapi = ctypes.CDLL("spinapi")
    except:
        print("Failed to load spinapi library.")
        pass


def enum(**enums):
    return type('Enum', (), enums)


ns = 1.0
us = 1000.0
ms = 1000000.0

unit_list = {
    'ns': 1.0,
    'us': 1000.0,
    'ms': 1000000.0
}

MHz = 1.0
kHz = 0.001
Hz = 0.000001

#   Instruction enum
Inst = enum(
    CONTINUE = 0,
    STOP = 1,
    LOOP = 2,
    END_LOOP = 3,
    JSR = 4,
    RTS = 5,
    BRANCH = 6,
    LONG_DELAY = 7,
    WAIT = 8,
    RTI = 9
)

inst_dict = {
    "CONTINUE": "0",
    "STOP": "1",
    "LOOP": "2",
    "END_LOOP": "3",
    "JSR": "4",
    "RTS": "5",
    "BRANCH": "6",
    "LONG_DELAY": "7",
    "WAIT": "8",
    "RTI": "9"
}

inst_list = [
    "CONTINUE",
    "STOP",
    "LOOP",
    "END_LOOP",
    "JSR",
    "RTS",
    "BRANCH",
    "LONG_DELAY",
    "WAIT",
    "RTI"
]


channel_list = {
    '0': 'scmos',
    '1': '780cooling',
    '2': '780repump',
    '3': 'EOM_ttl',
    '4': '795depump_back',
    '5': '795depump_front',
    '6': '795repump',
    '7': '795state_prepare',
    '8': '780probe_space_AOM',
    '9': 'reset',
    '10': 'single_photon_AOM',
    '11': 'v_dipole',
    '12': '480detune',
    '13': '480readout',
    '14': 'qutau_trigger',
    '15': '780probe_fiber_AOM',
    '16': 'spd_gate',
    '17': 'spd_gate',
    '18': 'spd_gate',
    '19': 'spd_gate',
    '20': 'unused',
    '21': 'always_on',
    '22': 'always_on',
    '23': 'always_on',
}

spinapi.pb_get_version.restype = (ctypes.c_char_p)
spinapi.pb_get_error.restype = (ctypes.c_char_p)

spinapi.pb_count_boards.restype = (ctypes.c_int)

spinapi.pb_init.restype = (ctypes.c_int)

spinapi.pb_select_board.argtype = (ctypes.c_int)
spinapi.pb_select_board.restype = (ctypes.c_int)

spinapi.pb_set_debug.argtype = (ctypes.c_int)
spinapi.pb_set_debug.restype = (ctypes.c_int)

spinapi.pb_set_defaults.restype = (ctypes.c_int)

spinapi.pb_core_clock.argtype = (ctypes.c_double)
spinapi.pb_core_clock.restype = (ctypes.c_int)

spinapi.pb_write_register.argtype = (ctypes.c_int, ctypes.c_int)
spinapi.pb_write_register.restype = (ctypes.c_int)

spinapi.pb_start_programming.argtype = (ctypes.c_int)
spinapi.pb_start_programming.restype = (ctypes.c_int)

spinapi.pb_stop_programming.restype = (ctypes.c_int)

spinapi.pb_start.restype = (ctypes.c_int)
spinapi.pb_stop.restype = (ctypes.c_int)
spinapi.pb_reset.restype = (ctypes.c_int)
spinapi.pb_close.restype = (ctypes.c_int)

spinapi.pb_inst_dds2.argtype = (
    ctypes.c_int,   # Frequency register DDS0
    ctypes.c_int,   # Phase register DDS0
    ctypes.c_int,   # Amplitude register DDS0
    ctypes.c_int,   # Output enable DDS0
    ctypes.c_int,   # Phase reset DDS0
    ctypes.c_int,   # Frequency register DDS1
    ctypes.c_int,   # Phase register DDS1
    ctypes.c_int,   # Amplitude register DDS1
    ctypes.c_int,   # Output enable DDS1,
    ctypes.c_int,   # Phase reset DDS1,
    ctypes.c_int,   # Flags
    ctypes.c_int,   # inst
    ctypes.c_int,   # inst data
    ctypes.c_double,    # timing value (double)
)

spinapi.pb_inst_dds2.restype = (ctypes.c_int)

spinapi.pb_inst_pbonly.argtype = (
    ctypes.c_int,   # Flags
    ctypes.c_int,   # inst
    ctypes.c_int,   # inst data
    ctypes.c_double,    # timing value (double)
)

spinapi.pb_inst_pbonly.restype = (ctypes.c_int)




def pb_get_version():
    """Return library version as UTF-8 encoded string."""
    ret = spinapi.pb_get_version()
    return str(ctypes.c_char_p(ret).value.decode("utf-8"))


def pb_get_error():
    """Return library error as UTF-8 encoded string."""
    ret = spinapi.pb_get_error()
    return str(ctypes.c_char_p(ret).value.decode("utf-8"))


def pb_count_boards():
    """Return the number of boards detected in the system."""
    return spinapi.pb_count_boards()


def pb_init():
	"""Initialize currently selected board."""
	return spinapi.pb_init()


def pb_set_debug(debug):
	return spinapi.pb_set_debug(debug)


def pb_select_board(board_number):
	"""Select a specific board number"""
	return spinapi.pb_select_board(board_number)


def pb_set_defaults():
	"""Set board defaults. Must be called before using any other board functions."""
	return spinapi.pb_set_defaults()


def pb_core_clock(clock):
	return spinapi.pb_core_clock(ctypes.c_double(clock))


def pb_write_register(address, value):
	return spinapi.pb_write_register(address, value)


def pb_start_programming(target):
	return spinapi.pb_start_programming(target)


def pb_stop_programming():
	return spinapi.pb_stop_programming()


def pb_inst_dds2(*args):
	t = list(args)
	# Argument 13 must be a double
	t[13] = ctypes.c_double(t[13])
	args = tuple(t)
	return spinapi.pb_inst_dds2(*args)


def pb_inst_pbonly(*args):
    t = list(args)
    t[3] = ctypes.c_double(t[3])
    args = tuple(t)
    return spinapi.pb_inst_pbonly(*args)


def pb_multi_inst(construction_list):
    for construction in construction_list:
        pb_inst_pbonly(construction)


def pb_start():
    return spinapi.pb_start()


def pb_stop():
    return spinapi.pb_stop()


def pb_reset():
    return spinapi.pb_reset()


def pb_close():
    return spinapi.pb_close()


def pb_test(Flags, inst, inst_data, timing):
    t = [Flags, inst, inst_data, timing]
    print(type(t)) #  Argument 4 must be a double
    t[3] = ctypes.c_double(t[3])
    success = 'You did it'
    return success





def get_pb_data(line):
    label = re.findall(r"^[a-zA-z]+", line)
    if len(label) != 0:
        label = label[0]
    else:
        label = ''

    output_pattern = re.findall(r"0b[\w\s]+",re.findall(r"0b[\w\s]+\,", line)[0])[0].replace(' ', '')  # 2进制
    # output_pattern = re.findall(r"0x[a-zA-Z0-9]+", line)[0].replace(' ','')  # 16进制
    time = re.findall(r"[0-9]+", re.findall(r"\,\s*[0-9]+", line)[0])[0]
    unit = re.findall(r"[a-z]s", re.findall(r"[0-9][a-z]s", line)[0])[0]

    inst = re.findall(r"\,\s*[a-zA-z]+\,*?", line)
    if len(inst) != 0:
        inst = inst_dict[re.findall(r"[a-zA-z]+", re.findall(r"\,\s*[a-zA-z]+\,*?", line)[0])[0]]
        inst_data = re.findall(r"\,\s*[0-9]+\s*//", line)
        if len(inst_data) != 0:
            inst_data = int(re.findall(r"[0-9]+", inst_data[0])[0])
        else:
            inst_data = 0
    else:
        inst = 0
        inst_data = 0
    # print(unit)
    # comment = re.findall(r"//.*", line)
    return output_pattern, time, inst, inst_data, unit

def pb_initail():
    pbclock=300        # 时钟频率300MHz

    # Enable the log file
    pb_set_debug(0)    # 0表示不进入编程状态

    print("Using SpinAPI Library version %s" % pb_get_version())         # 获得当前API版本，目前是20171214
    print("Found %d boards in the system.\n" % pb_count_boards())        # 检测当前的板子数量！

    pb_select_board(0)      # 0为默认值，是第一块板子

    if pb_init() != 0:  # 板子初始化 模式正常则返回0
        raise Exception("Number: %d" % pb_count_boards())
        # raise Exception("Error initializing board: %s" % pb_get_error())
        # print("Error initializing board: %s" % pb_get_error())     #返回错误类型
        # # input("Please press a key to continue.")
        # exit(-1)

    # Configure the core clock
    pb_core_clock(pbclock)  # 时钟频率设置


def load_channels(AllSignals, input_type = 'LoopedSignal'):
    """

    :param AllSignals: [[Port_1,Signal_1,Delay_1]....[port_n,Signal_n,Delay_n]]
    :return:
    """
    sequence_list = []
    delay_list = [AllSignals[i][2] for i in range(len(AllSignals))]
    max_delay = max(delay_list)
    for index, channel in enumerate(AllSignals):
        temp = ds.decode_multi_loop(channel[1])
        temp = ds.adjust_general_delay(temp, channel[2], max_delay)
        # print(index, temp)
        sequence_list.append([channel[0], temp])

    channels_sequence = ds.synchronize_complex_sqeuence(sequence_list)
    pb_reset()
    pb_start_programming(PULSE_PROGRAM)
    for index, sequence in enumerate(channels_sequence):
        # exec('pb_inst_pbonly(bin(int(sequence[0])), int(sequence[2]), int(sequence[3]), sequence[1])'.format{bin(int(sequence[0])),int(sequence[2]),int(sequence[3]),sequence[1]})
        # print(str(bin(int(sequence[0]))) + "," + str(int(sequence[1])) + "ns," + str(int(sequence[2])) + "," + str(
        #     int(sequence[3])))
        # print('pb_inst_pbonly({}, {}, {}, {})'
        #      .format(bin(int(sequence[0])), int(sequence[2]), int(sequence[3]), sequence[1]))


        if int(sequence[2]) == 0:
            print(str(bin(int(sequence[0])))+","+str(int(sequence[1]))+"ns")
        elif int(sequence[3]) == 0:
            print(str(bin(int(sequence[0]))) + "," + str(int(sequence[1])) + "ns," + inst_list[int(sequence[2])])
        else:
            print(str(bin(int(sequence[0])))+","+str(int(sequence[1]))+"ns,"+inst_list[int(sequence[2])]+","+str(int(sequence[3])))
    # pb_inst_pbonly(0b111000000000000000000000, Inst.CONTINUE, 0, 200.0 * us)
    pb_stop_programming()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import setuptools

with open("README.md", "r") as fh:
  long_description = fh.read()

setuptools.setup(
  name="PQRLab",
  version="0.4.0",
  author="PQR Team",
  author_email="carrollxiang@gmail.com",
  description="Controll System for PQR.",
  long_description="Controll System for PQR.",
  long_description_content_type="text/markdown",
  url="https://gitee.com/teshenghsiang/PQRLab",
  packages=setuptools.find_packages(),
  classifiers=[
  "Programming Language :: Python :: 3",
  "License :: OSI Approved :: MIT License",
  "Operating System :: OS Independent",
  ],
  include_package_data=True,
  package_data={'PQRLab': ['pqrcore/*']}
)


import numpy as np


import pqrspincore.DecodeSignal as ds


from pqrcore.experiment.signal import Signal, LoopedSignal


from pqrspincore.spinapi import *


unit_list = {
    'ns': 1.0,
    'us': 1000.0,
    'ms': 1000000.0
}

channel_list = {
    '0': 'scmos',
    '1': 'cooling_780',
    '2': 'repump_780',
    '3': 'EOM_ttl',
    '4': 'depump_back_795',
    '5': 'depump_front_795',
    '6': 'repump_795',
    '7': 'state_prepare_795',
    '8': 'probe_space_AOM_780',
    '9': 'reset',
    '10': 'single_photon_AOM',
    '11': 'v_dipole',
    '12': 'detune_480',
    '13': 'readout_480',
    '14': 'qutau_trigger',
    '15': 'probe_fiber_AOM_780',
    '16': 'spd_gate',
    '17': 'spd_gate',
    '18': 'spd_gate',
    '19': 'spd_gate',
    '20': 'unused',
    '21': 'always_on',
    '22': 'always_on',
    '23': 'always_on',
}

def new_get_pb_data(line):
    label = re.findall(r"^[a-zA-z]+", line)
    if len(label) != 0:
        label = label[0]
    else:
        label = ''

    output_pattern = re.findall(r"0b[\w\s]+", re.findall(r"0b[\w\s]+\,", line)[0])[0].replace(' ', '')  # 2进制
    # output_pattern = re.findall(r"0x[a-zA-Z0-9]+", line)[0].replace(' ','')  # 16进制
    time = re.findall(r"[0-9]+", re.findall(r"\,\s*[0-9]+", line)[0])[0]
    unit = re.findall(r"[a-z]s", re.findall(r"[0-9][a-z]s", line)[0])[0]

    inst = re.findall(r"\,\s*[a-zA-z]+\,*?", line)
    if len(inst) != 0:
        inst = inst_dict[re.findall(r"[a-zA-z]+", re.findall(r"\,\s*[a-zA-z]+\,*?", line)[0])[0]]
        inst_data = re.findall(r"\,\s*[0-9]+\s*//", line)
        if len(inst_data) != 0:
            inst_data = int(re.findall(r"[0-9]+", inst_data[0])[0])
        else:
            inst_data = 0
    else:
        inst = 0
        inst_data = 0
    # comment = re.findall(r"//.*", line)
    return output_pattern, time, inst, inst_data, unit

# str1 = 'mininal_time_sequence.pb'
# channels = [[] for i in range(24)]
#
# with open(str1, 'r+', encoding='UTF-8') as file:
#     for line in file:
#         if len(line.strip()) != 0:
#             pattern = new_get_pb_data(line)[0]
#             time = new_get_pb_data(line)[1]
#             inst = int(new_get_pb_data(line)[2])
#             inst_data = int(new_get_pb_data(line)[3])
#             unit = unit_list[new_get_pb_data(line)[4]]
#             for i in range(24):
#                 channels[i].append(np.array([int(pattern[i+2]), int(time)*unit, inst, inst_data]))
#
# for i in range(24):
#     channels[i] = np.vstack(channels[i])
#     channels[i] = ds.combine_similar_items(channels[i])
#     temp = []
#     for j in range(channels[i].shape[0]):
#         temp.append((str(channels[i][j][0]), channels[i][j][1]))
#     # print(temp)
#     channels[i] = temp
#     print(channel_list[str(23-i)], '=LoopedSignal(', channels[i], ',5)')
#
#
# LoopedSignal_list = [LoopedSignal(channels[i], 4)for i in range(24)]

# file = open('my.txt', 'w+')
# for i in range(len(data)):
#     s = str(data[i]).replace('[','').replace(']','')#去除[],这两行按数据不同，可以选择
#     s = s.replace("'",'').replace(',','') +'\n'   #去除单引号，逗号，每行末尾追加换行符
#     file.write(s)
# file.write(LoopedSignal_list)
# file.close()







# -*- coding: utf-8 -*-
"""
Created on 2020-04-10

@author: 张宽

实验信号处理
"""

from math import inf
from pqrcore.expression import isname, ParsedExpression


class Signal():
    """
    单通道连续信号
    self._sig - 连续信号 (ParsedExpression类)
    """

    def __init__(self, sig):
        self.signal = sig

    def __str__(self):
        return str(self._sig)

    def __call__(self, *times, **rules):
        """
        计算self.signal的实时电压
        times = [t_0, t_1, t_2, ...]
            t_0 - 从当前连续信号开始的时间
            t_1 - 从当前所在最内层循环开始的时间
            t_2 - 从更外一层循环开始的时间
            ...
        """
        i = 0
        for time in times:
            rules['t_%u' % i] = time
            i += 1
        return self._sig(**rules)

    @property
    def signal(self):
        return self._sig

    @signal.setter
    def signal(self, sig):
        if not isinstance(sig, ParsedExpression):
            sig = ParsedExpression(sig)
        self._sig = sig


class LoopedSignal(Signal):
    """
    单通道循环信号
    self._sig = [(sig, t), ...] - 循环内容
        sig - 当前信号从前一段的t开始 (Signal类)
        t - 当前sig结束时间，t=0为本次循环开始时刻 (float,str类)
    self.__loops - 循环次数 (int,str类或inf)
    """

    def __init__(self, sig, loops):
        self.signal = sig
        self.loops = loops

    def __str__(self):
        def str_with_indent(sig, indent):
            """将sig输出为字符串表达式，加缩进"""
            string = 'loops = %u:' % sig.__loops
            d = 0.
            for sig, t in sig._sig:
                string += '\n' + indent * '    '
                string += '%.3f <= t < %.3f, ' % (d, t)
                if isinstance(sig, LoopedSignal):
                    string += str_with_indent(sig, indent + 1)
                else:
                    string += str(sig)
                d = t
            return string

        return str_with_indent(self, 1)

    def __call__(self, t, *times, **rules):
        """
        计算self.signal的实时电压
        t - 从第一次循环开始的时间，通过t可以计算t_0, t_1, ..., t_i
        times = [t_i+1, t_i+2, ...]
        """
        loops = self.__loops
        if isinstance(loops, str):
            loops = rules[loops]
        sig = self._sig
        d = sig[-1][1]
        if isinstance(d, str):
            d = rules[d]
        if t >= loops * d:
            return 0
        t_i = t % d
        d = 0.
        for sig, t in sig:
            if isinstance(t, str):
                t = rules[t]
            if t > t_i:
                t = t_i - d
                break
            d = t
        return sig(t, t_i, *times, **rules)

    @property
    def signal(self):
        return self._sig

    @signal.setter
    def signal(self, sig):
        self._sig = []
        d = 0.
        for sig, t in sig:
            try:
                t = float(t)
            except ValueError as err:
                if not isname(t):
                    raise err
            else:
                if t <= d:
                    continue
                d = t
            if not isinstance(sig, Signal):
                sig = Signal(sig)
            self._sig.append((sig, t))

    @property
    def loops(self):
        return self.__loops

    @loops.setter
    def loops(self, loops):
        try:
            loops = int(loops)
        except OverflowError:
            loops = inf
        except ValueError as err:
            if not isname(loops):
                raise err
        else:
            if loops < 0:
                raise ValueError('循环信号循环次数不能<0')
        self.__loops = loops
